from types import SimpleNamespace


# TODO: make 'isEmpty' generic like the other utils


def isEmpty(lenAble):
    if isinstance(lenAble, SimpleNamespace):
        return len(lenAble.__dict__) is 0
    else:
        return len(lenAble) is 0



from inspect import signature
from ..internal.raise_ import raise_
import wrapt


def argIsInstance(aType, fnName=None):
    @wrapt.decorator
    def wrapper(fn, _instance, args, kwargs):
        nonlocal fnName

        if not isinstance(args[0], aType):
            argName = list(signature(fn).parameters)[0]
            fnName = fnName or fn.__name__
            typePassed = type(args[0])
            typeName = aType.__name__
            raise_(
                ValueError,
                f"""
                {fnName} requires {argName} to be an instance of {typeName}
                type passed: {typePassed.__name__}
                """,
            )

        return fn(*args, **kwargs)

    return wrapper



# ------- #
# Imports #
# ------- #

from types import SimpleNamespace
from .makeGenericCallFn import makeGenericCallFn
from .getTypedResult import getTypedResult
from ..decorators.argIsCallable import argIsCallable


# ---- #
# Main #
# ---- #


@argIsCallable
def map_(mapperFn):
    fnName = map_.__name__
    callMapperFn = makeGenericCallFn(mapperFn, 3, fnName)

    def map_inner(collection):
        typedMap = getTypedResult(collection, typeToMap, fnName)
        return typedMap(callMapperFn, collection)

    return map_inner


# ------- #
# Helpers #
# ------- #


def map_list(callMapperFn, aList):
    result = []
    for idx, el in enumerate(aList):
        result.append(callMapperFn(el, idx, aList))

    return result


def map_simpleNamespace(callMapperFn, aSimpleNamespace):
    result = SimpleNamespace()
    for key, val in aSimpleNamespace.__dict__.items():
        setattr(result, key, callMapperFn(val, key, aSimpleNamespace))

    return result


typeToMap = {list: map_list, SimpleNamespace: map_simpleNamespace}



from textwrap import dedent

#
# the 'grep' options are parsed in simple-test-process, but for sake of
#   usage formatting I'm placing their usage here.  This causes a potential
#   issue where the simple-test-process and this usage are out of sync,
#   but whatever.
#
usage = dedent(
    f"""
    Usage
      simple-test [options]
      simple-test (--help | --version)

    Options
      --project-dir  the project dir from which tests are found. Defaults
                     to `os.getcwd()`
      --silent       a flag which disables output and prevents the reporter from
                     being called.
      --reporter     module name with a 'report' function.  The default reporter
                     is 'simple_test_default_reporter'.  Relative modules e.g.
                     "..myModule" are not yet supported.
      --grep         filter which tests *and* suites are run
      --grep-tests   filter just the tests
      --grep-suites  filter just the suites

    Returns an exit code of
        0 for a successful run
        1 for a failed run
        2 for an error
    """
)



# ------- #
# Imports #
# ------- #

from .getTypedResult import getTypedResult


# ---- #
# Main #
# ---- #


def sort(collection):
    typedSort = getTypedResult(collection, typeToSort, "sort")
    return typedSort(collection)


# ------- #
# Helpers #
# ------- #


def sort_viaSorted(something):
    return sorted(something)


typeToSort = {list: sort_viaSorted}



from types import SimpleNamespace
from .mAssignToSelf import mAssignToSelf
from .raise_ import raise_
import os


class NotCallable(SimpleNamespace):
    def __init__(self, fnName, **typeToFn):
        self._fnName = fnName
        self._typeToFn = typeToFn
        mAssignToSelf(typeToFn, self)

    def __call__(self, *args, **kwargs):
        availableKeys = list(self._typeToFn.keys())
        fnName = self._fnName

        raise_(
            TypeError,
            f"""
            The utility '{fnName}' is not callable because it needs to know what
            to return in the case of an empty list.

            example usage:
            {fnName}.{availableKeys[0]}([...])

            available keys:
              {os.linesep.join(availableKeys)}
            """,
        )



# ------- #
# Imports #
# ------- #

from ordered_set import OrderedSet
from .getTypedResult import getTypedResult


# ---- #
# Main #
# ---- #


def joinWith(separator):
    def joinWith_inner(collection):
        typedJoinWith = getTypedResult(collection, typeToJoinWith, "joinWith")
        return typedJoinWith(separator, collection)

    return joinWith_inner


# ------- #
# Helpers #
# ------- #


def joinWith_iterable(separator, aList):
    return separator.join(aList)


typeToJoinWith = {list: joinWith_iterable, OrderedSet: joinWith_iterable}



#
# README
#  - I'm not sure what to call this.  Its purpose is to centralize the type
#    validation, minimizing copy/paste.
#

# ------- #
# Imports #
# ------- #

from .raise_ import raise_
from .get import get


# ---- #
# Init #
# ---- #

getName = get("__name__")

#
# we need a unique object here to ensure the 'typeToSomething' doesn't lead us
#   to a 'None' value
#
nothing = object()


# ---- #
# Main #
# ---- #


def getTypedResult(value, typeToSomething, fnName):
    valueType = type(value)
    result = typeToSomething.get(valueType, nothing)

    if _isSomething(result):
        return result

    supportedTypes = ", ".join(map(getName, typeToSomething.keys()))
    raise_(
        ValueError,
        f"""\
        {fnName} doesn't support the type '{valueType.__name__}'
        supported types: {supportedTypes}
        """,
    )


# ------- #
# Helpers #
# ------- #


def _isSomething(x):
    return x is not nothing



# ------- #
# Imports #
# ------- #

from case_conversion import camelcase
from copy import deepcopy
from simple_test_process.parseArgs import _grepArgs, _grepArgsKeys
from types import SimpleNamespace as o
from ..fns import isLaden
from .usage import usage
import os


# ---- #
# Init #
# ---- #

arguments = set(["--project-dir", "--reporter", "--silent", *_grepArgsKeys])
helpOrVersion = set(["--help", "--version"])


# ---- #
# Main #
# ---- #


def validateAndParseArgs(args, cliResult):
    argsObj = o(
        reporter=None,
        projectDir=None,
        silent=False,
        grepArgs=deepcopy(_grepArgs),
    )
    validationResult = o(
        argsObj=argsObj, cliResult=cliResult, hasError=False, positionalArgs=[]
    )

    i = 0

    while i < len(args):
        if not args[i].startswith("--"):
            break

        arg = args[i]
        if arg not in arguments:
            if not argsObj.silent:
                if arg in helpOrVersion:
                    cliResult.stderr = (
                        f"'{arg}' must be the only argument when passed"
                    )
                else:
                    cliResult.stderr = f"invalid option '{arg}'"
                    cliResult.stderr += os.linesep + usage

            cliResult.code = 2
            validationResult.hasError = True

            return validationResult

        if arg == "--silent":
            argsObj.silent = True

        elif arg == "--reporter":
            if i == len(args) - 1:
                if not argsObj.silent:
                    cliResult.stderr = "'--reporter' must be given a value"
                    cliResult.stderr += os.linesep + usage

                cliResult.code = 2
                validationResult.hasError = True
                return validationResult

            i += 1
            arg = args[i]
            argsObj.reporter = arg

        elif arg == "--project-dir":
            i += 1
            if i == len(args):
                if not argsObj.silent:
                    cliResult.stderr = "'--project-dir' must be given a value"
                    cliResult.stderr += os.linesep + usage

                cliResult.code = 2
                validationResult.hasError = True
                return validationResult

            arg = args[i]
            argsObj.projectDir = arg

        else:  # arg in _grepArgsKeys
            i += 1
            if i == len(args):
                if not argsObj.silent:
                    cliResult.stderr = f"{arg} must be given a value"
                    cliResult.stderr += os.linesep + usage

                cliResult.code = 2
                validationResult.hasError = True
                return validationResult

            grepVals = getattr(argsObj.grepArgs, camelcase(arg))
            grepVals.append(args[i])

        i += 1

    positionalArgs = args[i:]
    if isLaden(positionalArgs):
        if not argsObj.silent:
            cliResult.stderr = "this command doesn't take positional arguments"
            cliResult.stderr += os.linesep + usage

        cliResult.code = 2
        validationResult.hasError = True
        return validationResult

    return validationResult



# ------- #
# Imports #
# ------- #

from math import inf
from .returnFirstArgument import returnFirstArgument as identity
from .getNumPositionalParams import getNumPositionalParams


# ---- #
# Main #
# ---- #


def makeCallFn(fn, callerName, *, modifyResult=identity):
    (_, allowed) = getNumPositionalParams(fn, callerName)

    if allowed is inf:
        return lambda *args, **kwargs: modifyResult(fn(*args, **kwargs))

    else:
        return lambda *args, **kwargs: modifyResult(
            fn(*args[:allowed], **kwargs)
        )



# ------- #
# Imports #
# ------- #

from os import path
from simple_test_process.parseArgs import _grepArgs
from types import SimpleNamespace
from ..fns import all_, isEmpty, isSomething, raise_


# ---- #
# Init #
# ---- #

setOfGrepArgsKeys = set(_grepArgs.__dict__.keys())


# ---- #
# Main #
# ---- #


def validateRunParams(grepArgs, projectDir, reporter, silent):
    if not isinstance(grepArgs, SimpleNamespace):
        raise_(
            TypeError,
            f"""
            'grepArgs' must be an instance of SimpleNamespace
            type: {type(grepArgs).__name__}
            """,
        )

    if not hasAllGrepKeys(grepArgs):
        raise_(
            ValueError,
            f"""
            'grepArgs' must contain all (and only) the available keys
            available keys: {", ".join(_grepArgs.__dict__.keys())}
            keys given: {", ".join(grepArgs.__dict__.keys())}
            """,
        )

    if not all_(areStringLists)(grepArgs):
        raise ValueError("'grepArgs' can only contain lists of strings")

    if isSomething(projectDir):
        if not isinstance(projectDir, str):
            raise_(
                TypeError,
                f"""\
                'projectDir' must be an instance of str
                type: {type(silent).__name__}
                """,
            )

        if isEmpty(projectDir):
            raise ValueError("'projectDir' cannot be an empty string")

        if not path.isabs(projectDir):
            raise_(
                ValueError,
                f"""
                'projectDir' must pass 'os.path.isabs'
                projectDir: {projectDir}
                """,
            )

        if not path.isdir(projectDir):
            raise_(
                ValueError,
                f"""
                'projectDir' must pass 'os.path.isdir'
                projectDir: {projectDir}
                """,
            )

    if isSomething(reporter):
        if not isinstance(reporter, str):
            raise_(
                TypeError,
                f"""
                'reporter' must be an instance of str
                type: {type(reporter).__name__}
                """,
            )

        if isEmpty(reporter):
            raise ValueError("'reporter' cannot be an empty string")

        if reporter.startswith("."):
            raise_(
                ValueError,
                f"""
                relative reporter modules are not yet supported
                reporter: {reporter}
                """,
            )

    if isSomething(silent) and not isinstance(silent, bool):
        raise_(
            TypeError,
            f"""\
            'silent' must be an instance of bool
            type: {type(silent).__name__}
            """,
        )


# ------- #
# Helpers #
# ------- #


def hasAllGrepKeys(grepArgs):
    if len(grepArgs.__dict__) != len(_grepArgs.__dict__):
        return False

    for key in grepArgs.__dict__.keys():
        if key not in setOfGrepArgsKeys:
            return False

    return True


def areStrings(val):
    return isinstance(val, str)


def areStringLists(val):
    return isinstance(val, list) and all_(areStrings)(val)



from inspect import signature
from .raise_ import raise_


def getFnSignature(fn, callerName):
    try:
        return signature(fn)
    except Exception as e:
        raise_(
            ValueError,
            f"""\
            '{callerName}' is unable to get the signature of the passed callable
            callable passed: {fn.__name__}

            one reason this could occur is the callable is written in c (e.g.
            the builtin 'str' callable).
            """,
            fromException=e,
        )



# ------- #
# Imports #
# ------- #

from case_conversion import dashcase
from copy import deepcopy
from os import path
from simple_test_process.parseArgs import _grepArgs
from .validateRunParams import validateRunParams
import os
import sys

from ..fns import (
    flattenDeep,
    getListOfCollectionValues,
    isLaden,
    keepWhen,
    map_,
    passThrough,
    raise_,
)


# ---- #
# Main #
# ---- #


def createRun(subprocessRun):
    return lambda **kwargs: run(subprocessRun, **kwargs)


def run(
    subprocessRun,
    *,
    grepArgs=None,
    projectDir=None,
    reporter=None,
    silent=False,
):
    if grepArgs is None:
        grepArgs = deepcopy(_grepArgs)

    validateRunParams(grepArgs, projectDir, reporter, silent)

    if projectDir is None:
        projectDir = os.getcwd()
    else:
        projectDir = path.normpath(projectDir)

    if reporter is None:
        reporter = "simple_test_default_reporter"

    ensureTestsDirExists(projectDir)

    cliGrepArgs = toCliGrepArgs(grepArgs)

    subprocessResult = subprocessRun(
        [
            sys.executable,
            "-m",
            "simple_test_process",
            projectDir,
            reporter,
            str(silent),
            *cliGrepArgs,
        ],
        cwd=projectDir,
    )

    return subprocessResult.returncode


# ------- #
# Helpers #
# ------- #


def ensureTestsDirExists(projectDir):
    testsDir = path.join(projectDir, "tests")
    if not path.isdir(testsDir):
        raise_(
            Exception,
            f"""
            projectDir must contain a directory 'tests'
            projectDir: {projectDir}
            """,
        )


def eachToKeyValuePair(grepVals, grepKey):
    cliGrepKey = f"--{dashcase(grepKey)}"
    return map_(lambda val: [cliGrepKey, val])(grepVals)


def toCliGrepArgs(grepArgs):
    return passThrough(
        grepArgs,
        [
            keepWhen(isLaden),
            map_(eachToKeyValuePair),
            getListOfCollectionValues,
            flattenDeep,
        ],
    )



from .raise_ import raise_

ascDescSet = {"asc", "desc"}


def sanitizeAscDesc(ascOrDesc):
    sanitized = ascOrDesc.lower()
    if sanitized not in ascDescSet:
        raise_(
            ValueError,
            f"""
            ascOrDesc must be either 'asc' or 'desc' (case insensitive)
            value given: {ascOrDesc}
            """,
        )

    return sanitized



from .cli import runSimpleTest
from .run import run
import sys


def printErr(msg):
    print(msg, file=sys.stderr)


def main():
    result = runSimpleTest(run, sys.argv[1:])

    if result.stdout:
        print(result.stdout)

    if result.stderr:
        printErr(result.stderr)

    exit(result.code)



# ------- #
# Imports #
# ------- #

from types import SimpleNamespace
from .internal.makeGenericCallFn import makeGenericCallFn
from .internal.getTypedResult import getTypedResult
from .decorators.argIsCallable import argIsCallable


# ---- #
# Main #
# ---- #


@argIsCallable
def all_(predicate):
    fnName = all.__name__
    callPredicate = makeGenericCallFn(predicate, 3, fnName)

    def all_inner(collection):
        typedAll = getTypedResult(collection, typeToAll, fnName)
        return typedAll(callPredicate, collection)

    return all_inner


# ------- #
# Helpers #
# ------- #


def all_list(callPredicate, aList):
    for idx, el in enumerate(aList):
        if not callPredicate(el, idx, aList):
            return False

    return True


def all_simpleNamespace(callPredicate, aSimpleNamespace):
    for key, val in aSimpleNamespace.__dict__.items():
        if not callPredicate(val, key, aSimpleNamespace):
            return False

    return True


typeToAll = {list: all_list, SimpleNamespace: all_simpleNamespace}



# ------- #
# Imports #
# ------- #

from math import inf
from .getNumPositionalParams import getNumPositionalParams
from .raise_ import raise_


# ---- #
# Main #
# ---- #


def makeGenericCallFn(fn, maxParams, callerName):
    required, allowed = getNumPositionalParams(fn, callerName)

    if allowed is inf:
        allowed = maxParams

    if required > maxParams:
        raise_(
            ValueError,
            f"""
            {callerName} can only take functions with up to {maxParams}
            positional params.  The function '{fn.__name__}' requires {required}
            """,
        )

    return lambda *args, **kwargs: fn(*args[:allowed], **kwargs)



# ------- #
# Imports #
# ------- #

from types import SimpleNamespace
from .internal.makeGenericCallFn import makeGenericCallFn
from .internal.getTypedResult import getTypedResult
from .decorators.argIsCallable import argIsCallable


# ---- #
# Main #
# ---- #


@argIsCallable
def forEach(fn):
    fnName = forEach.__name__
    callFn = makeGenericCallFn(fn, 3, fnName)

    def forEach_inner(collection):
        typedForEach = getTypedResult(collection, typeToForEach, fnName)
        return typedForEach(callFn, collection)

    return forEach_inner


# ------- #
# Helpers #
# ------- #


def forEach_dict(callFn, aDict):
    for key, val in aDict.items():
        callFn(val, key, aDict)

    return aDict


def forEach_list(callFn, aList):
    for idx, el in enumerate(aList):
        callFn(el, idx, aList)

    return aList


def forEach_simpleNamespace(callFn, aSimpleNamespace):
    forEach_dict(callFn, aSimpleNamespace.__dict__)
    return aSimpleNamespace


typeToForEach = {
    dict: forEach_dict,
    list: forEach_list,
    SimpleNamespace: forEach_simpleNamespace,
}



from inspect import isclass, signature
import wrapt


@wrapt.decorator
def argIsClass(fn, _instance, args, kwargs):
    if not isclass(args[0]):
        argName = list(signature(fn).parameters)[0]
        fnName = fn.__name__
        raise ValueError(f"{fnName} requires {argName} to be a class")

    return fn(*args, **kwargs)



# ------- #
# Imports #
# ------- #

from .getTypedResult import getTypedResult
from .makeGenericCallFn import makeGenericCallFn
from ..decorators.argIsCallable import argIsCallable


# ---- #
# Main #
# ---- #


@argIsCallable
def discardWhen(predicate):
    fnName = discardWhen.__name__
    shouldDiscard = makeGenericCallFn(predicate, 3, fnName)

    def discardWhen_inner(collection):
        typedDiscardWhen = getTypedResult(collection, typeToDiscardWhen, fnName)
        return typedDiscardWhen(shouldDiscard, collection)

    return discardWhen_inner


# ------- #
# Helpers #
# ------- #


def discardWhen_list(shouldDiscard, aList):
    result = []
    for idx, el in enumerate(aList):
        if not shouldDiscard(el, idx, aList):
            result.append(el)

    return result


def discardWhen_dict(shouldDiscard, aDict):
    result = {}
    for key, val in aDict.items():
        if shouldDiscard(val, key, aDict):
            result[key] = val

    return result


typeToDiscardWhen = {list: discardWhen_list, dict: discardWhen_dict}



from inspect import isclass


def isType(aType):
    if not isclass(aType):
        raise ValueError("isType requires argument 'aType' to pass isclass")

    def isType_inner(something):
        return type(something) is aType

    return isType_inner



# ------- #
# Imports #
# ------- #

from types import SimpleNamespace
from .internal.getTypedResult import getTypedResult


# ---- #
# Main #
# ---- #


def getListOfCollectionValues(collection):
    fnName = getListOfCollectionValues.__name__
    typedValues = getTypedResult(collection, typeToValues, fnName)
    return typedValues(collection)


# ------- #
# Helpers #
# ------- #


def getListOfCollectionValues_dict(aDict):
    return list(aDict.values())


def getListOfCollectionValues_simpleNamespace(aSimpleNamespace):
    return list(aSimpleNamespace.__dict__.values())


typeToValues = {
    dict: getListOfCollectionValues_dict,
    SimpleNamespace: getListOfCollectionValues_simpleNamespace,
}



from ordered_set import OrderedSet
import wrapt

from ..internal.discardWhen import discardWhen
from ..internal.get import get
from ..internal.getArgName import getArgName
from ..internal.isLaden import isLaden
from ..internal.isType import isType
from ..internal.joinWith import joinWith
from ..internal.map_ import map_
from ..internal.passThrough import passThrough
from ..internal.raise_ import raise_
from ..internal.sort import sort
from ..internal.toType import toType


def argIsListOfType(aType):
    @wrapt.decorator
    def wrapper(fn, _instance, args, kwargs):
        typePassed = type(args[0])

        fnName = fn.__name__
        typeName = aType.__name__

        if typePassed is not list:
            argName = getArgName(fn)
            raise_(
                ValueError,
                f"""\
                {fnName} requires {argName} to have the type list
                type passed: {typePassed.__name__}
                """,
            )

        invalidTypes = discardWhen(isType(aType))(args[0])
        if isLaden(invalidTypes):
            argName = getArgName(fn)
            invalidTypeNames = passThrough(
                invalidTypes,
                [
                    map_(toType),
                    OrderedSet,
                    list,
                    map_(get("__name__")),
                    sort,
                    joinWith(", "),
                ],
            )
            raise_(
                ValueError,
                f"""\
                {fnName} requires {argName} to be a list of {typeName}
                invalid types passed: {invalidTypeNames}
                """,
            )

        return fn(*args, **kwargs)

    return wrapper



from inspect import signature
import wrapt


@wrapt.decorator
def argIsCallable(fn, _instance, args, kwargs):
    if not callable(args[0]):
        argName = list(signature(fn).parameters)[0]
        fnName = fn.__name__
        raise ValueError(f"{fnName} requires {argName} to be callable")

    return fn(*args, **kwargs)



from tedent import tedent
from .isOnlyWhitespace import isOnlyWhitespace
import os


def raise_(errorClass, message, *, fromException=None):
    allLines = message.split(os.linesep)
    if isOnlyWhitespace(allLines[0]) and isOnlyWhitespace(allLines[-1]):
        message = tedent(message)

    err = errorClass(message)

    if fromException is None:
        raise err
    else:
        raise err from fromException



# ------- #
# Imports #
# ------- #

from types import SimpleNamespace
from .internal.makeGenericCallFn import makeGenericCallFn
from .internal.getTypedResult import getTypedResult
from .decorators.argIsCallable import argIsCallable


# ---- #
# Main #
# ---- #


@argIsCallable
def keepWhen(predicate):
    fnName = keepWhen.__name__
    shouldKeep = makeGenericCallFn(predicate, 3, fnName)

    def keepWhen_inner(collection):
        typedKeepWhen = getTypedResult(collection, typeToKeepWhen, fnName)
        return typedKeepWhen(shouldKeep, collection)

    return keepWhen_inner


# ------- #
# Helpers #
# ------- #


def keepWhen_list(shouldKeep, aList):
    result = []
    for idx, el in enumerate(aList):
        if shouldKeep(el, idx, aList):
            result.append(el)

    return result


def keepWhen_dict(shouldKeep, aDict):
    result = {}
    for key, val in aDict.items():
        if shouldKeep(val, key, aDict):
            result[key] = val

    return result


def keepWhen_simpleNamespace(shouldKeep, aSimpleNamespace):
    result = SimpleNamespace()
    for key, val in aSimpleNamespace.__dict__.items():
        if shouldKeep(val, key, aSimpleNamespace):
            setattr(result, key, val)

    return result


typeToKeepWhen = {
    list: keepWhen_list,
    dict: keepWhen_dict,
    SimpleNamespace: keepWhen_simpleNamespace,
}



# ------- #
# Imports #
# ------- #

from .makeGenericCallFn import makeGenericCallFn
from .getTypedResult import getTypedResult
from ..decorators.argIsCallable import argIsCallable


# ---- #
# Main #
# ---- #


@argIsCallable
def reduce(fn, initial):
    reducerFn = makeGenericCallFn(fn, 4, "reduce")

    def reduce_inner(collection):
        typedReduce = getTypedResult(collection, typeToReduce, "reduce")
        return typedReduce(reducerFn, initial, collection)

    return reduce_inner


# ------- #
# Helpers #
# ------- #


def reduce_list(reducerFn, initial, aList):
    result = initial
    for idx, el in enumerate(aList):
        result = reducerFn(result, el, idx, aList)

    return result


typeToReduce = {list: reduce_list}



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['po',
 'po.simple_test',
 'po.simple_test.cli',
 'po.simple_test.fns',
 'po.simple_test.fns.decorators',
 'po.simple_test.fns.internal',
 'po.simple_test.run']

package_data = \
{'': ['*']}

install_requires = \
['case_conversion>=2.1,<3.0',
 'ordered_set>=3.1,<4.0',
 'simple_test_default_reporter>=0.2.0,<0.3.0',
 'simple_test_process>=0.2.0,<0.3.0',
 'tedent>=0.1.1,<0.2.0',
 'wrapt>=1.10,<2.0']

entry_points = \
{'console_scripts': ['simple-test = po.simple_test.script:main']}

setup_kwargs = {
    'name': 'po.simple-test',
    'version': '0.2.0',
    'description': 'A simple test runner',
    'long_description': 'I will document this module if people will find it helpful.  Currently\nsimple_test is just for my own personal use.\n',
    'author': 'Philip Olson',
    'author_email': 'philip.olson@pm.me',
    'url': 'https://github.com/olsonpm/py_simple-test',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



#
# assigns dict key-values onto self as attributes
# return secondaryObj
#   ** mutates secondaryObj
#


def mAssignToSelf(aDict, aSelf):
    for k, v in aDict.items():
        setattr(aSelf, k, v)

    return aSelf



from .all_ import all_
from .flattenDeep import flattenDeep
from .forEach import forEach
from .getListOfCollectionValues import getListOfCollectionValues
from .invokeAttr import invokeAttr
from .isEmpty import isEmpty
from .isLaden import isLaden
from .isSomething import isSomething
from .joinWith import joinWith
from .justReturn import justReturn
from .keepWhen import keepWhen
from .map_ import map_
from .passThrough import passThrough
from .raise_ import raise_

__all__ = [
    "all_",
    "flattenDeep",
    "forEach",
    "getListOfCollectionValues",
    "invokeAttr",
    "isEmpty",
    "isLaden",
    "isSomething",
    "joinWith",
    "justReturn",
    "keepWhen",
    "map_",
    "passThrough",
    "raise_",
]



# ------- #
# Imports #
# ------- #

from inspect import Parameter
from math import inf
from .getFnSignature import getFnSignature
from .iif import iif


# ---- #
# Init #
# ---- #

_nonVarPositionalParamKinds = {
    Parameter.POSITIONAL_ONLY,
    Parameter.POSITIONAL_OR_KEYWORD,
}


# ---- #
# Main #
# ---- #

#
# Returns a tuple
#   numRequired: number of required positional params
#   numAllowed:  number of allowed positional params
#


def getNumPositionalParams(fn, callerName):
    sig = getFnSignature(fn, callerName)

    numAllowed = iif(_hasVarPositionalParam(sig), inf, 0)
    numRequired = 0

    for p in sig.parameters.values():
        if p.kind in _nonVarPositionalParamKinds:
            numAllowed += 1
            if p.default is Parameter.empty:
                numRequired += 1

    return (numRequired, numAllowed)


# ------- #
# Helpers #
# ------- #


def _hasVarPositionalParam(sig):
    for p in sig.parameters.values():
        if p.kind is Parameter.VAR_POSITIONAL:
            return True



# ------- #
# Imports #
# ------- #

from .internal.getTypedResult import getTypedResult


# ---- #
# Main #
# ---- #


def flattenDeep(collection):
    fnName = flattenDeep.__name__
    typedFlattenDeep = getTypedResult(collection, typeToFlattenDeep, fnName)
    return typedFlattenDeep(collection)


# ------- #
# Helpers #
# ------- #


def flattenDeep_list(aList):
    result = []

    def flattenDeep_recursive(aList):
        for maybeInnerList in aList:
            if isinstance(maybeInnerList, list):
                flattenDeep_recursive(maybeInnerList)
            else:
                result.append(maybeInnerList)

    flattenDeep_recursive(aList)

    return result


typeToFlattenDeep = {list: flattenDeep_list}



# ------- #
# Imports #
# ------- #

from traceback import format_exc
from types import SimpleNamespace as o
from ..meta import version
from .usage import usage
from .validateAndParseArgs import validateAndParseArgs
import os


# ---- #
# Main #
# ---- #


def createRunSimpleTest(run):
    return lambda args: runSimpleTest(run, args)


def runSimpleTest(run, args):
    result = o(stdout=None, stderr=None, code=None)

    numArgs = len(args)
    if numArgs == 1:
        if args[0] == "--help":
            result.stdout = usage
            result.code = 0
            return result
        elif args[0] == "--version":
            result.stdout = version
            result.code = 0
            return result

    validationResult = validateAndParseArgs(args, result)

    if validationResult.hasError:
        return validationResult.cliResult

    argsObj = validationResult.argsObj
    isSilent = argsObj.silent

    try:
        subprocessReturnCode = run(
            grepArgs=argsObj.grepArgs,
            reporter=argsObj.reporter,
            projectDir=argsObj.projectDir,
            silent=isSilent,
        )

        result.code = subprocessReturnCode
        return result

    except:
        if not isSilent:
            result.stderr = (
                "An unexpected error occurred" + (os.linesep * 2) + format_exc()
            )

        result.code = 2
        return result
